function [K_all, Inv_all] = QuadrotorSDEllipsoidalKernels_v2(method,L0_num,plot_results,...
                                        uWeightFactor,CheckForInvariance,...
                                        computeKfree,useNAHSalg)
% 
% [K_all, Inv_all] = QuadrotorSDEllipsoidalKernels(method,L0_num,plot_results,use_fix,CheckForInvariance,computeKfree,useNAHSalg)
% 
% Computes the less conservative sampled data discriminating kernel for the quadrotor
% example using ellipsoidal techniques. Improves upon the NAHS algorithm.
%
% 
% INPUTS:   method              Indicates the approximation method used: 
%                               1 = under-approximate K_k x U, compute Inv, project;
%                               2 = over-approximate K_k x U, compute Inv over-approximation,
%                               optimize, project (default). Recommended.
% 
%           L0_num              Number of initial directions -- the first 9 are fixed, 
%                               rest random. (Default 20)
% 
%           plot_results        Binary flag to plot 2D projection results. (Default 1)
% 
%           uWeightFactor       Weighting factor in [0,1] that dictates how
%                               much the input dimensions should matter when
%                               computing the maximum volume ellipsoids in the
%                               augmented space. 0 means that all emphasis
%                               is put on maximizing the size of the
%                               discriminating kernels and thus may result in
%                               smaller input sets. (Default 0.01)
%                               For best results consider taking the union of 
%                               the sets in x-space and u-space generated by multiple
%                               values of this parameter.
% 
%           CheckForInvariance  Check if the invariance condition holds. (Default 0)
% 
%           computeKfree        Compute K_free? (Default 0)
% 
%           useNAHSalg          Use the NAHS algorithm? If 0 (default), the new tight 
%                               algorithm is used.
% 
% OUTPUTS:  K_all               Cell array: each element corresponds to a direction and 
%                               contains all intermediate discriminating kernels K_k. 
%                               Its last element for a fixed l (i.e. K_all{l}(end)) is K_free 
%                               and its first element is K_0.
% 
%           Inv_all             Cell array: contains the computed intermediate invariance 
%                               kernels Inv_k(K_k x U) which can be used
%                               during synthesis. If method=2, Inv_all
%                               contains sets that *over-approximate* Inv_k
%                               in the direction of U. For synthesis purposes,
%                               intersect their projection onto u-space with U.
% 
% 
% S. Kaynama, last modified: 9/10/14, for HSCC'15 submission. kaynama@ece.ubc.ca


if nargin < 7, useNAHSalg = 0; end
if nargin < 6, computeKfree = 0; end
if nargin < 5, CheckForInvariance = 0; end
if nargin < 4, uWeightFactor = 0.01; end
if nargin < 3, plot_results = 1; end
if nargin < 2, L0_num = 20; end
if nargin < 1, method = 2; end


gravity = 9.81;
uhover = 8.96; %Thrust value needed for hover (empirically measured); see Kene's email (flagged)

kT = gravity/(uhover^2);
kp = 1/0.15;

xeq = [2 0 sqrt(gravity/kT)]';
ueq = sqrt(gravity/kT); %(The same as uhover.)

a = [0 1 0;
    0 0 2*kT*xeq(3);
    0 0 -kp];

b = [0 0 kp]';

g = [0 1 0]'; 

n = size(a,1);
m = size(b,2);

%Control set \U
uMin = 0;
uMax = 10;
uLinMax = uMax-ueq;
uLinMin = uMin-ueq;
U = (uLinMax-uLinMin)/2*ell_unitball(m)+(uLinMax+uLinMin)/2;

%Disturbance set \V
vMax = .1; %Affects x2 only (in +ve direction), depends on value of x3 { CHANGE }
vMin = 0;
V = ellipsoid((vMax+vMin)/2*ones(1,1),((vMax-vMin)/2)^2*eye(1));

%Constraint set \K
ceiling = 2.8; floor = 0.5; %meters
acclMin = 8; acclMax = 10; %bounds on acceleration x3, determines disturbance on x2 { CHANGE } 
velMin = -1.5; velMax = 1.5;
K0 = ellipsoid(-xeq + [(ceiling+floor)/2; (velMin+velMax)/2 ;(acclMin+acclMax)/2],...
    diag([ ((ceiling-floor)/2)^2 ((velMax-velMin)/2)^2 ((acclMax-acclMin)/2)^2 ]));

d = 1/10; %Sampling time at 10-30 Hz.
T = [0 1]; %Horizon.
N = T(2)/d;
InvRes = d/200; %Sub-sample step length (i.e. resolution in Inv calculation)


% %%%%%%%%%%%%%%%%%%%%%%%
% %Uncomment for the Oscillating Double Integrator (NAHS example)
% a = [ 0 -10 0; 10 0 0; 2 2 0 ];
% b = [ 1; 0; 0 ];
% g = [0 0 0]';
% n = size(a,1);
% m = size(b,2);
% K0 = ell_unitball(n);
% U = ell_unitball(m);
% V = ell_unitball(m);
% T = [0 3];
% d = .1;
% InvRes = .0033;
% N = T(2)/d;
% %%%%%%%%%%%%%%%%%%%%%%%


%Augmented dynamics
A = [a b; zeros(m,n+m)]; 
G = [g; zeros(m,size(g,2))];

lsys = linsys(A, zeros(n+m,m), U, G, V);

global ellOptions
ellOptions.verbose = 0;
ellOptions.plot2d_grid = 500;
ellOptions.time_grid = 4;

%Generate terminal directions:
L_stack = generateTerminalDirections(n,m,4,[1 2; 1 3; 2 3]);

lcount = 0;

%~~~~~~~~~~~~~~~~~~~ Main Algorithm ~~~~~~~~~~~~~~~~~~~~~~
tic
for l = 1:L0_num,

    disp(['--------Direction #' num2str(l) '--------'])
    L0 = L_stack(:,l)

    BadDirectionFlag = 0;
    
    clear K Inv
    K(1:N+1) = ellipsoid;
    Inv(1:N+1) = ellipsoid;
    K(1) = K0;
    
    createKsubset = 1; %Flag to create a subset of target for Invg().
    
    for k = 1:N,

        disp(['Iteration ' num2str(k)])
        
        [K(k+1), Inv(k+1)] = Disc1(K(k),Inv(1),[0 d],L0,lsys,method,...
            InvRes,createKsubset,CheckForInvariance,uWeightFactor);
        if k == 1 && ~useNAHSalg, Inv(k)=Inv(k+1); end  %Basically forces [] to be passed for Inv(1), which reverts things back to NAHS.
        
        createKsubset = 0; %Stop eroding for next iterations.
        
        if isempty(K(k+1)) || (volume(K(k+1))<1e-7),
            disp(['Empty Disc approximtion at the end of iteration k=' num2str(k)])
            BadDirectionFlag = 1;
            break;
        end
        
    end
    
    %Formulate K_free
    if computeKfree && ~isempty(K(k+1)) && (volume(K(k+1))>1e-7) && ~BadDirectionFlag,
        disp('Computing K_{free}...')
        [Uq, UQ] = double(U);
        [Vq, VQ] = double(V);
        Eu = ellipsoid([Uq; Vq], [UQ  zeros(m,m); zeros(m,m) 10^2*VQ]); %U x R^m
        Ev = ellipsoid([Uq; Vq], [10^2*UQ  zeros(m,m); zeros(m,m) VQ]); %R^m x V
        UxV = intersection_ea(Eu, Ev); %Over-approximate UxV w/ minvol enclosing ellipsoid.
        lsys_f_wc = linsys(a, b, U, [b g], UxV);
        L0_n = any(L0(1:n)~=0)*L0(1:n) + ~any(L0(1:n)~=0)*2*rand(n,1)-1;
        if ~useNAHSalg,
            K(k+2) = Invg(K(k+1),projection(Inv(1),[eye(n); zeros(m,n)]),[0 d],...
                L0_n,lsys_f_wc,InvRes,CheckForInvariance,uWeightFactor);
        else
            K(k+2) = Invg(K(k+1),Inv(1),[0 d],...
                L0_n,lsys_f_wc,InvRes,CheckForInvariance,uWeightFactor);
        end
        [Uq, UQ] = double(U);
        [K0q, K0Q] = double(K0);
        Inv(k+2) = ellipsoid([K0q; Uq], [20^2*K0Q zeros(n,m); zeros(m,n) UQ]); %R^n x U;        
    else
        K(k+2) = ellipsoid;
        Inv(k+2) = ellipsoid;
    end

    %Save the sets (only if the end disc approximate is non-empty)
    if ~BadDirectionFlag, %all(~isempty(K(1:end-1))),
        lcount = lcount+1;
        K_all{lcount} = K;
        Inv_all{lcount} = Inv;
    end
    
end
elapsed_time = toc
%~~~~~~~~~~~~~~~~ End of Main Algorithm ~~~~~~~~~~~~~~~~~~~~


%Plot 2D projection of results
if plot_results,
    figure('position',[0,0,1000,300]);
    movegui(gcf,'center')
    options.fill = 1;
    I = eye(n);
    proj = [1 2; 1 3; 2 3];
    %names = {'$z$', '$\dot{z}$', '$\ddot{z}$'};
    names = {'$\bar{x}_1$', '$\bar{x}_2$', '$\bar{x}_3$'};
    for i = 1:3,
        subplot(1,3,i)
        plot(projection(K0,I(:,proj(i,:))),'r',options)
        hold on
        for l = 1:lcount,
            plot(projection(K_all{l}(end-1),I(:,proj(i,:))),'g',options)
            if ~isempty(K_all{l}(end)),
                plot(projection(K_all{l}(end),I(:,proj(i,:))),'b',options)
            end
        end
        xlabel(names{proj(i,1)},'Interpreter','Latex')
        ylabel(names{proj(i,2)},'Interpreter','Latex')
        axis square
        box on; grid
        set(get(gca,'xlabel'),'fontsize',16)
        set(get(gca,'ylabel'),'fontsize',16)
    end
end

end


function [unitstepDisc, Ed] = Disc1(K,Inv1,Tsub,L0,lsys,method,InvRes,createKsubset,CheckForInvariance,uWeightFactor)
% Implements $Disc_1$. Depending on METHOD, either under-approximates,
% computes, and projects, or over-approximates, computes, optimizes,
% and projects

nplusm = size(lsys.A,1);
m = dimension(lsys.control);
n = nplusm - m;
I = eye(nplusm);

%Erode the constraint set K?
if createKsubset,
    disp('Eroding the target...')
    ll = 2*rand(n,100)-1;
    KK = [lsys.A(1:n,1:n)*K  lsys.A(1:n,n+1:n+m)*lsys.control  lsys.G(1:n,:)*lsys.disturbance]; 
    KA = minksum_ea(KK, ll);
    [~,minvol_ind] = min(volume(KA));
    MP = minkdiff_ia(K, InvRes*KA(minvol_ind), ll);
    [~,maxvol_ind] = max(volume(MP));
    K = MP(maxvol_ind);
end

[Kq, KQ] = double(K);
[Uq, UQ] = double(lsys.control);


Ex = ellipsoid([Kq; Uq], [KQ  zeros(n,m); ...
    zeros(m,n) 4^2*UQ]); %K x R^m. Also over-approximates K x U for method 2.

Eu = ellipsoid([Kq; Uq], [10^2*KQ  zeros(n,m); ...
    zeros(m,n) UQ]); %R^n x U


switch method

    case 1 %under-approximate Kk x U, compute Inv, project for Disc

        Ed = Invg(ellintersection_ia([Ex Eu]),Inv1,Tsub,L0,lsys,InvRes,CheckForInvariance,uWeightFactor);
        %For max trace use ellIntersectionIA instead.
        
        if isempty(Ed),
            unitstepDisc = ellipsoid;
            return
        end

        unitstepDisc = projection(Ed,I(:,1:n));

    case 2 %over-approximate Kk x U, compute Inv (over-approximation), optimize, project

        Ed = Invg(Ex,Inv1,Tsub,L0,lsys,InvRes,CheckForInvariance,uWeightFactor);

        if isempty(Ed),
            unitstepDisc = ellipsoid;
            return
        end

        %Find ellipsoid in Inv x U whose x-space projection has largest
        %volume. Project to get Disc under-approximation.
        P = [eye(n) zeros(n,m)];
        Ez = ellIntersectionIAProjET([Ed Eu],P);
        unitstepDisc = projection(Ez,I(:,1:n));
        
end

end


function Inv1K = Invg(E,Inv1,Tsub,L0,lsys,InvRes,CheckForInvariance,uWeightFactor)
% Computes ellipsoidal approximation of $Inv_1$ over TSUB in the
% direction L0. For new algorith it does: First iteration via subinterval 
% recusion, the rest using Reach(K_k x U) \cap Inv_1(K x U).

options.approximation = 1;  %0: external, 1:internal, 2:both

N = round(Tsub(2)/InvRes); %number of sub-intervals
Pt = Tsub(2):-abs(Tsub(2)-Tsub(1))/N:Tsub(1); %Setup interval partition

nplusm = size(lsys.A,1);
m = dimension(lsys.control);
n = nplusm - m;


if uWeightFactor < 0 || uWeightFactor > 1,
    error('uWeightFactor should be between 0 and 1!');
end
Px = [eye(n) zeros(n,m)];
Pu = [zeros(m,n) eye(m)];


if isempty(Inv1),

    %For the first iteration where Inv1 is still empty:
    new_tgt = E;
    
    disp('Computing inv kernel in R^{n+m} (subinterval recursions)...')
    for i = 1:length(Pt)-1,
        %Recursive algorithm for fixed L0    
        rs = reach(lsys, new_tgt, L0, [abs(diff(Pt(i:i+1))) 0], options); %The backward reach set at ti-ti_1
        
        rsE = get_ia(rs); %Get the internal approximating ellpses.
        reachset_ellps = rsE(:,end); %Save the final reach set.
        
        %Intersect with original target (or its erosion)
        new_tgt = ellIntersectionIAProjET_tradeoff([E  reachset_ellps],Px,Pu,uWeightFactor);
        
        if isempty(new_tgt) || (volume(new_tgt)<1e-7),
            disp(['INVG: Intersection with target subset was empty (or almost empty) at the end of sub-interval #'...
                num2str(i) ' for the current direction.'])
            break
        end
        
    end
    Inv1K = new_tgt;

else
    
    %After the first iteration:
    rs = reach(lsys, E, L0, [Tsub(2) 0], options);
        
    rsE = get_ia(rs); %Get the internal approximating ellpses
    reachset_ellps = rsE(:,end);
     
    Inv1K = ellIntersectionIAProjET_tradeoff([reachset_ellps  Inv1],Px,Pu,uWeightFactor); %Intersect with Inv1 (or Disc1 really)
    
    if CheckForInvariance && isinside(reachset_ellps,E),
        disp('Controlled-invariance condition is satisifed!')
    end

end

end


function L_stack = generateTerminalDirections(n,m,num,subspc)
% Generates a stack of terminal unit-norm directions, in this order: 
% - Normalized vector of 1s in R^{n+m};
% - Standard basis vectors in R^{n+m};
% - Uniformly partitioned vectors in [0, pi] in 2D subspaces
%   specified by rows of subspc. 
%   Total number depends on num: if num is even, generates num
%   vectors in each subspace including the standard bases; otherwise, 
%   generates num+1 vectors including the standard bases.
% - Randomly generated (uniform distrobution) vectors in R^{n+m}.
% 
% Total number of returned unit-norm vectors is 1+num (or 1+num+1 if num is
% odd) fixed vectors, plus a large number of random vectors.

theta = linspace(0,1*pi, num+1);
theta = theta(1:end-1);
for j=1:num
    L1(:,j) = [cos(theta(j)) -sin(theta(j));
        sin(theta(j))  cos(theta(j))] * [1;0];
end

L_stackfixed = zeros(n+m, size(subspc,1)*num);
for k_stackloop=1:size(subspc,1)
    L_stackfixed( subspc(k_stackloop,:), (k_stackloop-1)*num+1:k_stackloop*num) = L1;
end
L_stackfixed = L_stackfixed(:,all(abs(L_stackfixed)~=1)); %exclude the axses (which would include repetitions)
L_stackfixed = [eye(n+m), L_stackfixed]; %<--- Include normal axes?

L0_numrand = 200; %Some large number. For random directions.
L_stackrand = randn(n+m,L0_numrand);
L_stackrand = bsxfun(@rdivide,L_stackrand,sqrt(sum(L_stackrand.^2,2))); %Normalize.
L_stack = [ones(n+m,1)/sqrt(n+m), L_stackfixed, L_stackrand];  
    
end
